from typing import Union

import numpy as np

import jesse.helpers as jh
import jesse.services.selectors as selectors
from jesse.config import config
from jesse.enums import trade_types, order_types
from jesse.exceptions import EmptyPosition, OpenPositionError
from jesse.models import Order, Exchange
from jesse.services import logger, notifier
from jesse.utils import sum_floats, subtract_floats


class Position:
    def __init__(self, exchange_name: str, symbol: str, attributes: dict = None) -> None:
        self.id = jh.generate_unique_id()
        self.entry_price = None
        self.exit_price = None
        self.current_price = None
        self.qty = 0
        self.opened_at = None
        self.closed_at = None
        self._mark_price = None
        self._funding_rate = None
        self._next_funding_timestamp = None
        self._liquidation_price = None

        if attributes is None:
            attributes = {}

        self.exchange_name = exchange_name
        self.exchange: Exchange = selectors.get_exchange(self.exchange_name)

        self.symbol = symbol
        self.strategy = None

        for a in attributes:
            setattr(self, a, attributes[a])

    @property
    def mark_price(self) -> float:
        if not jh.is_live():
            return self.current_price

        return self._mark_price

    @property
    def funding_rate(self) -> float:
        if not jh.is_live():
            return 0

        return self._funding_rate

    @property
    def next_funding_timestamp(self) -> Union[int, None]:
        if not jh.is_live():
            return None

        return self._next_funding_timestamp

    @property
    def value(self) -> float:
        """
        The value of open position in the quote currency

        :return: float
        """
        return abs(self.current_price * self.qty)

    @property
    def type(self) -> str:
        """
        The type of open position - long, short, or close

        :return: str
        """
        if self.qty > 0:
            return 'long'
        if self.qty < 0:
            return 'short'

        return 'close'

    @property
    def pnl_percentage(self) -> float:
        """
        Alias for self.roi

        :return: float
        """
        return self.roi

    @property
    def roi(self) -> float:
        """
        Return on Investment in percentage
        More at: https://www.binance.com/en/support/faq/5b9ad93cb4854f5990b9fb97c03cfbeb
        """
        return self.pnl / self.total_cost * 100

    @property
    def total_cost(self) -> float:
        """
        How much we paid to open this position (currently does not include fees, should we?!)
        """
        if self.is_close:
            return np.nan

        base_cost = self.entry_price * abs(self.qty)
        if self.strategy:
            return base_cost / self.leverage

        return base_cost

    @property
    def leverage(self) -> Union[int, np.float64]:
        if self.exchange.type == 'spot':
            return 1

        if self.strategy:
            return self.strategy.leverage
        else:
            return np.nan

    @property
    def entry_margin(self) -> float:
        """
        Alias for self.total_cost
        """
        return self.total_cost

    @property
    def pnl(self) -> float:
        """
        The PNL of the position

        :return: float
        """
        if self.qty == 0:
            return 0

        diff = self.value - abs(self.entry_price * self.qty)

        return -diff if self.type == 'short' else diff

    @property
    def is_open(self) -> bool:
        """
        Is the current position open?

        :return: bool
        """
        return self.qty != 0

    @property
    def is_close(self) -> bool:
        """
        Is the current position close?

        :return: bool
        """
        return self.qty == 0

    @property
    def mode(self) -> str:
        if self.exchange.type == 'spot':
            return 'spot'
        else:
            return self.exchange.futures_leverage_mode

    @property
    def liquidation_price(self) -> Union[float, np.float64]:
        """
        The price at which the position gets liquidated. formulas are taken from:
        https://help.bybit.com/hc/en-us/articles/900000181046-Liquidation-Price-USDT-Contract-
        """
        if self.is_close:
            return np.nan

        if jh.is_livetrading():
            return self._liquidation_price

        if self.mode in ['cross', 'spot']:
            return np.nan

        elif self.mode == 'isolated':
            if self.type == 'long':
                return self.entry_price * (1 - self._initial_margin_rate + 0.004)
            elif self.type == 'short':
                return self.entry_price * (1 + self._initial_margin_rate - 0.004)
            else:
                return np.nan

        else:
            raise ValueError

    @property
    def _initial_margin_rate(self) -> float:
        return 1 / self.leverage

    @property
    def bankruptcy_price(self) -> Union[float, np.float64]:
        if self.type == 'long':
            return self.entry_price * (1 - self._initial_margin_rate)
        elif self.type == 'short':
            return self.entry_price * (1 + self._initial_margin_rate)
        else:
            return np.nan

    @property
    def to_dict(self):
        return {
            'entry_price': self.entry_price,
            'qty': self.qty,
            'current_price': self.current_price,
            'value': self.value,
            'type': self.type,
            'exchange': self.exchange_name,
            'pnl': self.pnl,
            'pnl_percentage': self.pnl_percentage,
            'leverage': self.leverage,
            'liquidation_price': self.liquidation_price,
            'bankruptcy_price': self.bankruptcy_price,
            'mode': self.mode,
        }

    def _close(self, close_price: float) -> None:
        if self.is_open is False:
            raise EmptyPosition('The position is already closed.')

        # just to prevent confusion
        close_qty = abs(self.qty)

        estimated_profit = jh.estimate_PNL(
            close_qty, self.entry_price,
            close_price, self.type
        )
        entry = self.entry_price
        trade_type = self.type
        self.exit_price = close_price

        if self.exchange:
            self.exchange.add_realized_pnl(estimated_profit)
            self.exchange.temp_reduced_amount[jh.base_asset(self.symbol)] += abs(close_qty * close_price)
        self.qty = 0
        self.entry_price = None
        self.closed_at = jh.now_to_timestamp()

        if not jh.is_unit_testing():
            info_text = f'CLOSED {trade_type} position: {self.exchange_name}, {self.symbol}, {self.strategy.name}. PNL: ${round(estimated_profit, 2)}, Balance: ${jh.format_currency(round(self.exchange.wallet_balance(self.symbol), 2))}, entry: {entry}, exit: {close_price}'

            if jh.is_debuggable('position_closed'):
                logger.info(info_text)

            if jh.is_live() and config['env']['notifications']['events']['updated_position']:
                notifier.notify(info_text)

    def _reduce(self, qty: float, price: float) -> None:
        if self.is_open is False:
            raise EmptyPosition('The position is closed.')

        # just to prevent confusion
        qty = abs(qty)

        estimated_profit = jh.estimate_PNL(qty, self.entry_price, price, self.type)

        if self.exchange:
            # self.exchange.increase_futures_balance(qty * self.entry_price + estimated_profit)
            self.exchange.add_realized_pnl(estimated_profit)
            self.exchange.temp_reduced_amount[jh.base_asset(self.symbol)] += abs(qty * price)

        if self.type == trade_types.LONG:
            self.qty = subtract_floats(self.qty, qty)
        elif self.type == trade_types.SHORT:
            self.qty = sum_floats(self.qty, qty)

        info_text = f'REDUCED position: {self.exchange_name}, {self.symbol}, {self.type}, {self.qty}, ${round(self.entry_price, 2)}'

        if jh.is_debuggable('position_reduced'):
            logger.info(info_text)

        if jh.is_live() and config['env']['notifications']['events']['updated_position']:
            notifier.notify(info_text)

    def _increase(self, qty: float, price: float) -> None:
        if not self.is_open:
            raise OpenPositionError('position must be already open in order to increase its size')

        qty = abs(qty)
        # size = qty * price

        # if self.exchange:
        #     self.exchange.decrease_futures_balance(size)

        self.entry_price = jh.estimate_average_price(qty, price, self.qty,
                                                     self.entry_price)

        if self.type == trade_types.LONG:
            self.qty = sum_floats(self.qty, qty)
        elif self.type == trade_types.SHORT:
            self.qty = subtract_floats(self.qty, qty)

        info_text = f'INCREASED position: {self.exchange_name}, {self.symbol}, {self.type}, {self.qty}, ${round(self.entry_price, 2)}'

        if jh.is_debuggable('position_increased'):
            logger.info(info_text)

        if jh.is_live() and config['env']['notifications']['events']['updated_position']:
            notifier.notify(info_text)

    def _open(self, qty: float, price: float, change_balance: bool = True) -> None:
        if self.is_open:
            raise OpenPositionError('an already open position cannot be opened')

        self.entry_price = price
        self.exit_price = None
        self.qty = qty
        self.opened_at = jh.now_to_timestamp()

        info_text = f'OPENED {self.type} position: {self.exchange_name}, {self.symbol}, {self.qty}, ${round(self.entry_price, 2)}'

        if jh.is_debuggable('position_opened'):
            logger.info(info_text)

        if jh.is_live() and config['env']['notifications']['events']['updated_position']:
            notifier.notify(info_text)

    def _on_executed_order(self, order: Order) -> None:
        qty = order.qty
        price = order.price

        # TODO: detect reduce_only order, and if so, see if you need to adjust qty and price (above variables)

        self.exchange.charge_fee(qty * price)

        # order opens position
        if self.qty == 0:
            change_balance = order.type == order_types.MARKET
            self._open(qty, price, change_balance)
        # order closes position
        elif (sum_floats(self.qty, qty)) == 0:
            self._close(price)
        # order increases the size of the position
        elif self.qty * qty > 0:
            if order.is_reduce_only:
                logger.info('Did not increase position because order is a reduce_only order')
            else:
                self._increase(qty, price)
        # order reduces the size of the position
        elif self.qty * qty < 0:
            # if size of the order is big enough to both close the
            # position AND open it on the opposite side
            if abs(qty) > abs(self.qty):
                if order.is_reduce_only:
                    logger.info(
                        f'Executed order is bigger than the current position size but it is a reduce_only order so it just closes it. Order QTY: {qty}, Position QTY: {self.qty}')
                    self._close(price)
                else:
                    logger.info(
                        f'Executed order is big enough to not close, but flip the position type. Order QTY: {qty}, Position QTY: {self.qty}')
                    diff_qty = sum_floats(self.qty, qty)
                    self._close(price)
                    self._open(diff_qty, price)
            else:
                self._reduce(qty, price)

        if self.strategy:
            self.strategy._on_updated_position(order)



from collections import namedtuple

import numpy as np
import talib

from jesse.helpers import get_candle_source
from jesse.helpers import slice_candles

KST = namedtuple('KST', ['line', 'signal'])


def kst(candles: np.ndarray, sma_period1: int = 10, sma_period2: int = 10, sma_period3: int = 10, sma_period4: int = 15,
        roc_period1: int = 10, roc_period2: int = 15, roc_period3: int = 20, roc_period4: int = 30,
        signal_period: int = 9, source_type: str = "close", sequential: bool = False) -> KST:
    """
    Know Sure Thing (KST)

    :param candles: np.ndarray
    :param sma_period1: int - default: 10
    :param sma_period2: int - default: 10
    :param sma_period3: int - default: 10
    :param sma_period4: int - default: 15
    :param roc_period1: int - default: 10
    :param roc_period2: int - default: 15
    :param roc_period3: int - default: 20
    :param roc_period4: int - default: 30
    :param signal_period: int - default: 9
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: KST(line, signal)
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)

    aroc1 = talib.SMA(talib.ROC(source, timeperiod=roc_period1), sma_period1)
    aroc2 = talib.SMA(talib.ROC(source, timeperiod=roc_period2), sma_period2)
    aroc3 = talib.SMA(talib.ROC(source, timeperiod=roc_period3), sma_period3)
    aroc4 = talib.SMA(talib.ROC(source, timeperiod=roc_period4), sma_period4)
    line = aroc1[aroc1.size - aroc4.size:] + 2 * aroc2[aroc2.size - aroc4.size:] + \
           3 * aroc3[aroc3.size - aroc4.size:] + 4 * aroc4
    signal = talib.SMA(line, signal_period)

    if sequential:
        return KST(line, signal)
    else:
        return KST(line[-1], signal[-1])



import requests

import jesse.helpers as jh
from jesse import exceptions
from jesse.modes.import_candles_mode.drivers.interface import CandleExchange
from typing import Union

class BybitPerpetual(CandleExchange):
    def __init__(self) -> None:
        # import here instead of the top of the file to prevent possible the circular imports issue
        from jesse.modes.import_candles_mode.drivers.binance import Binance

        super().__init__(
            name='Bybit Perpetual',
            count=200,
            rate_limit_per_second=4,
            backup_exchange_class=Binance
        )

        self.endpoint = 'https://api.bybit.com/public/linear/kline'

    def get_starting_time(self, symbol: str) -> int:
        dashless_symbol = jh.dashless_symbol(symbol)

        payload = {
            'interval': 'W',
            'symbol': dashless_symbol,
            'limit': 200,
            'from': 1514811660
        }

        response = requests.get(self.endpoint, params=payload)

        # Exchange In Maintenance
        if response.status_code == 502:
            raise exceptions.ExchangeInMaintenance('ERROR: 502 Bad Gateway. Please try again later')

        # unsupported symbol
        if response.status_code == 400:
            raise ValueError(response.json()['msg'])

        if response.status_code != 200:
            raise Exception(response.content)

        data = response.json()['result']

        # since the first timestamp doesn't include all the 1m
        # candles, let's start since the second day then
        return int(data[1]['open_time']) * 1000

    def fetch(self, symbol: str, start_timestamp: int) -> Union[list, None]:
        dashless_symbol = jh.dashless_symbol(symbol)

        payload = {
            'interval': 1,
            'symbol': dashless_symbol,
            'from': int(start_timestamp / 1000),
            'limit': self.count,
        }

        response = requests.get(self.endpoint, params=payload)

        # Exchange In Maintenance
        if response.status_code == 502:
            raise exceptions.ExchangeInMaintenance('ERROR: 502 Bad Gateway. Please try again later')

        # unsupported symbol
        if response.status_code == 400:
            raise ValueError(response.json()['msg'])

        if response.status_code != 200:
            return

        data = response.json()['result']

        return [{
            'id': jh.generate_unique_id(),
            'symbol': symbol,
            'exchange': self.name,
            'timestamp': int(d['open_time']) * 1000,
            'open': float(d['open']),
            'close': float(d['close']),
            'high': float(d['high']),
            'low': float(d['low']),
            'volume': float(d['volume'])
        } for d in data]



from typing import Union

import numpy as np
try:
    from numba import njit
except ImportError:
    njit = lambda a : a

from jesse.helpers import get_candle_source, slice_candles


def rsx(candles: np.ndarray, period: int = 14, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    Relative Strength Xtra (rsx)
   
    :param candles: np.ndarray
    :param period: int - default: 14
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)
    res = rsx_fast(source, period)

    return res if sequential else res[-1]


@njit
def rsx_fast(source, period):
    # variables
    f0 = 0
    f8 = 0
    # f10 = 0
    f18 = 0
    f20 = 0
    f28 = 0
    f30 = 0
    f38 = 0
    f40 = 0
    f48 = 0
    f50 = 0
    f58 = 0
    f60 = 0
    f68 = 0
    f70 = 0
    f78 = 0
    f80 = 0
    f88 = 0
    f90 = 0

    # v4 = 0
    # v8 = 0
    # v10 = 0
    v14 = 0
    # v18 = 0
    v20 = 0

    # vC = 0
    # v1C = 0

    res = np.full_like(source, np.nan)

    for i in range(period, source.size):
        if f90 == 0:
            f90 = 1.0
            f0 = 0.0
            f88 = period - 1.0 if period >= 6 else 5.0
            f8 = 100.0 * source[i]
            f18 = 3.0 / (period + 2.0)
            f20 = 1.0 - f18
        else:
            f90 = f88 + 1 if f88 <= f90 else f90 + 1
            f10 = f8
            f8 = 100 * source[i]
            v8 = f8 - f10
            f28 = f20 * f28 + f18 * v8
            f30 = f18 * f28 + f20 * f30
            vC = f28 * 1.5 - f30 * 0.5
            f38 = f20 * f38 + f18 * vC
            f40 = f18 * f38 + f20 * f40
            v10 = f38 * 1.5 - f40 * 0.5
            f48 = f20 * f48 + f18 * v10
            f50 = f18 * f48 + f20 * f50
            v14 = f48 * 1.5 - f50 * 0.5
            f58 = f20 * f58 + f18 * abs(v8)
            f60 = f18 * f58 + f20 * f60
            v18 = f58 * 1.5 - f60 * 0.5
            f68 = f20 * f68 + f18 * v18
            f70 = f18 * f68 + f20 * f70
            v1C = f68 * 1.5 - f70 * 0.5
            f78 = f20 * f78 + f18 * v1C
            f80 = f18 * f78 + f20 * f80
            v20 = f78 * 1.5 - f80 * 0.5
            if f88 >= f90 and f8 != f10:
                f0 = 1.0
            if f88 == f90 and f0 == 0.0:
                f90 = 0.0
        if f88 < f90 and v20 > 0.0000000001:
            v4 = (v14 / v20 + 1.0) * 50.0
            v4 = min(v4, 100.0)
            v4 = max(v4, 0.0)
        else:
            v4 = 50.0
        res[i] = v4
    return res



import jesse.helpers as jh
import jesse.services.logger as logger
from jesse.enums import sides, order_types
from jesse.exceptions import NegativeBalance, InvalidConfig
from jesse.models import Order
from .Exchange import Exchange


class SpotExchange(Exchange):
    def add_realized_pnl(self, realized_pnl: float) -> None:
        pass

    def charge_fee(self, amount: float) -> None:
        pass

    # current holding assets
    assets = {}
    # current available assets (dynamically changes based on active orders)
    available_assets = {}

    def __init__(self, name: str, starting_assets: list, fee_rate: float):
        super().__init__(name, starting_assets, fee_rate, 'spot')

        from jesse.routes import router
        # check if base assets are configured
        for route in router.routes:
            base_asset = jh.base_asset(route.symbol)
            if base_asset not in self.available_assets:
                raise InvalidConfig(
                    f"Jesse needs to know the balance of your base asset for spot mode. Please add {base_asset} to your exchanges assets config.")

    def wallet_balance(self, symbol: str = '') -> float:
        if symbol == '':
            raise ValueError
        quote_asset = jh.quote_asset(symbol)
        return self.assets[quote_asset]

    def available_margin(self, symbol: str = '') -> float:
        return self.wallet_balance(symbol)

    def on_order_submission(self, order: Order, skip_market_order=True):
        base_asset = jh.base_asset(order.symbol)
        quote_asset = jh.quote_asset(order.symbol)

        # skip market order at the time of submission because we don't have
        # the exact order.price. Instead, we call on_order_submission() one
        # more time at time of execution without "skip_market_order=False".
        if order.type == order_types.MARKET and skip_market_order:
            return

        # used for logging balance change
        temp_old_quote_available_asset = self.available_assets[quote_asset]
        temp_old_base_available_asset = self.available_assets[base_asset]

        if order.side == sides.BUY:
            quote_balance = self.available_assets[quote_asset]
            self.available_assets[quote_asset] -= (abs(order.qty) * order.price) * (1 + self.fee_rate)
            if self.available_assets[quote_asset] < 0:
                raise NegativeBalance(
                    f"Balance cannot go below zero in spot market. Available capital at {self.name} for {quote_asset} is {quote_balance} but you're trying to sell {abs(order.qty * order.price)}"
                )
        # sell order
        else:
            base_balance = self.available_assets[base_asset]
            new_base_balance = base_balance + order.qty
            if new_base_balance < 0:
                raise NegativeBalance(
                    f"Balance cannot go below zero in spot market. Available capital at {self.name} for {base_asset} is {base_balance} but you're trying to sell {abs(order.qty)}"
                )

            self.available_assets[base_asset] -= abs(order.qty)

        temp_new_quote_available_asset = self.available_assets[quote_asset]
        if jh.is_debuggable('balance_update') and temp_old_quote_available_asset != temp_new_quote_available_asset:
            logger.info(
                f'Available balance for {quote_asset} on {self.name} changed from {round(temp_old_quote_available_asset, 2)} to {round(temp_new_quote_available_asset, 2)}'
            )
        temp_new_base_available_asset = self.available_assets[base_asset]
        if jh.is_debuggable('balance_update') and temp_old_base_available_asset != temp_new_base_available_asset:
            logger.info(
                f'Available balance for {base_asset} on {self.name} changed from {round(temp_old_base_available_asset, 2)} to {round(temp_new_base_available_asset, 2)}'
            )

    def on_order_execution(self, order: Order) -> None:
        base_asset = jh.base_asset(order.symbol)
        quote_asset = jh.quote_asset(order.symbol)

        if order.type == order_types.MARKET:
            self.on_order_submission(order, skip_market_order=False)

        # used for logging balance change
        temp_old_quote_asset = self.assets[quote_asset]
        temp_old_quote_available_asset = self.available_assets[quote_asset]
        temp_old_base_asset = self.assets[base_asset]
        temp_old_base_available_asset = self.available_assets[base_asset]

        # works for both buy and sell orders (sell order's qty < 0)
        self.assets[base_asset] += order.qty

        if order.side == sides.BUY:
            self.available_assets[base_asset] += order.qty
            self.assets[quote_asset] -= (abs(order.qty) * order.price) * (1 + self.fee_rate)
        # sell order
        else:
            self.available_assets[quote_asset] += abs(order.qty) * order.price * (1 - self.fee_rate)
            self.assets[quote_asset] += abs(order.qty) * order.price * (1 - self.fee_rate)

        temp_new_quote_asset = self.assets[quote_asset]
        if jh.is_debuggable('balance_update') and temp_old_quote_asset != temp_new_quote_asset:
            logger.info(
                f'Balance for {quote_asset} on {self.name} changed from {round(temp_old_quote_asset, 2)} to {round(temp_new_quote_asset, 2)}'
            )
        temp_new_quote_available_asset = self.available_assets[quote_asset]
        if jh.is_debuggable('balance_update') and temp_old_quote_available_asset != temp_new_quote_available_asset:
            logger.info(
                f'Balance for {quote_asset} on {self.name} changed from {round(temp_old_quote_available_asset, 2)} to {round(temp_new_quote_available_asset, 2)}'
            )

        temp_new_base_asset = self.assets[base_asset]
        if jh.is_debuggable('balance_update') and temp_old_base_asset != temp_new_base_asset:
            logger.info(
                f'Balance for {base_asset} on {self.name} changed from {round(temp_old_base_asset, 2)} to {round(temp_new_base_asset, 2)}'
            )
        temp_new_base_available_asset = self.available_assets[base_asset]
        if jh.is_debuggable('balance_update') and temp_old_base_available_asset != temp_new_base_available_asset:
            logger.info(
                f'Balance for {base_asset} on {self.name} changed from {round(temp_old_base_available_asset, 2)} to {round(temp_new_base_available_asset, 2)}'
            )

    def on_order_cancellation(self, order: Order) -> None:
        base_asset = jh.base_asset(order.symbol)
        quote_asset = jh.quote_asset(order.symbol)

        # used for logging balance change
        temp_old_quote_available_asset = self.available_assets[quote_asset]
        temp_old_base_available_asset = self.available_assets[base_asset]

        if order.side == sides.BUY:
            self.available_assets[quote_asset] += (abs(order.qty) * order.price) * (1 + self.fee_rate)
        # sell order
        else:
            self.available_assets[base_asset] += abs(order.qty)

        temp_new_quote_available_asset = self.available_assets[quote_asset]
        if jh.is_debuggable('balance_update') and temp_old_quote_available_asset != temp_new_quote_available_asset:
            logger.info(
                f'Available balance for {quote_asset} on {self.name} changed from {round(temp_old_quote_available_asset, 2)} to {round(temp_new_quote_available_asset, 2)}'
            )
        temp_new_base_available_asset = self.available_assets[base_asset]
        if jh.is_debuggable('balance_update') and temp_old_base_available_asset != temp_new_base_available_asset:
            logger.info(
                f'Available balance for {base_asset} on {self.name} changed from {round(temp_old_base_available_asset, 2)} to {round(temp_new_base_available_asset, 2)}'
            )



from jesse import utils
from jesse.strategies import Strategy


class TestFuturesExchangeAvailableMargin(Strategy):
    def before(self):
        if self.index == 0:
            assert self.position.exchange.wallet_balance() == 10000 == self.capital == self.balance
            assert self.position.exchange.available_margin() == 20000 == self.available_margin

        if self.price == 11:
            # wallet balance should have stayed the same while we haven't spent from it yet
            assert self.position.exchange.wallet_balance() == 10000 == self.balance
            # but available_margin should have
            assert round(self.position.exchange.available_margin()) == 20000 - 2000 == round(self.available_margin)

        if self.price == 12:
            # wallet balance should have changed because of fees, but we have fee=0 in this test, so:
            assert self.position.exchange.wallet_balance() == 10000 == self.balance
            # position is opened at this price, but the available_margin should have stayed the same (and NOT
            # include the take_profit and stop_loss which are reduce only orders
            assert round(self.position.exchange.available_margin()) == 20000 - 2000 + self.position.pnl == round(self.available_margin)

        if self.price == 13:
            # wallet balance should have changed because of fees, but we have fee=0 in this test, so:
            assert self.position.exchange.wallet_balance() == 10000 == self.balance
            # position is opened at this price, but the available_margin should have stayed the same (and NOT
            # include the take_profit and stop_loss which are reduce only orders. NOTICE that the unrealized
            # profit of the open position multiplied by the leverage should also have been added
            assert round(self.available_margin) == round(20000 - 2000 + self.position.pnl * self.leverage)

        if self.price == 21:
            # wallet balance must now equal to 10_000 + the profit we made from previous trade
            previous_trade = self.trades[0]
            assert self.position.exchange.wallet_balance() == previous_trade.pnl + 10000
            # now that position is closed, available_margin should equal to wallet_balance*leverage
            assert self.available_margin == self.position.exchange.wallet_balance() * 2
            assert self.balance == self.position.exchange.wallet_balance()
            assert self.balance * self.leverage == self.available_margin

    def should_long(self) -> bool:
        return self.price == 10

    def should_short(self) -> bool:
        return False

    def go_long(self):
        qty = utils.size_to_qty(2000, 12)
        self.buy = qty, 12
        self.take_profit = qty, 20
        self.stop_loss = qty, 10

    def go_short(self):
        pass

    def should_cancel(self):
        return False



from typing import Union

import numpy as np
import talib

from jesse.helpers import slice_candles


def sarext(candles: np.ndarray, start_value: float = 0, offset_on_reverse: float = 0, acceleration_init_long: float = 0,
           acceleration_long: float = 0, acceleration_max_long: float = 0, acceleration_init_short: float = 0,
           acceleration_short: float = 0, acceleration_max_short: float = 0, sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    SAREXT - Parabolic SAR - Extended

    :param candles: np.ndarray
    :param start_value: float - default: 0
    :param offset_on_reverse: float - default: 0
    :param acceleration_init_long: float - default: 0
    :param acceleration_long: float - default: 0
    :param acceleration_max_long: float - default: 0
    :param acceleration_init_short: float - default: 0
    :param acceleration_short: float - default: 0
    :param acceleration_max_short: float - default: 0
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    res = talib.SAREXT(candles[:, 3], candles[:, 4], startvalue=start_value, offsetonreverse=offset_on_reverse,
                       accelerationinitlong=acceleration_init_long, accelerationlong=acceleration_long,
                       accelerationmaxlong=acceleration_max_long, accelerationinitshort=acceleration_init_short,
                       accelerationshort=acceleration_short, accelerationmaxshort=acceleration_max_short)

    return res if sequential else res[-1]



from jesse.strategies import Strategy


# test_before
class TestBeforeMethod(Strategy):
    def should_long(self) -> bool:
        if self.index == 0:
            assert self.vars['counter'] == 10

        elif self.index == 2:
            assert self.vars['counter'] == 100

        return False

    def should_short(self) -> bool:
        return False

    def go_long(self):
        pass

    def go_short(self):
        pass

    def should_cancel(self):
        return False

    def before(self):
        if self.index == 0:
            self.vars['counter'] = 10

        elif self.index == 2:
            self.vars['counter'] = 100



from abc import ABC, abstractmethod


class CandleExchange(ABC):
    def __init__(self, name: str, count: int, rate_limit_per_second: float, backup_exchange_class):
        self.name = name
        self.count = count
        self.sleep_time = 1 / rate_limit_per_second
        self._backup_exchange_class = backup_exchange_class
        self._backup_exchange = None

    @property
    def backup_exchange(self):
        if self._backup_exchange_class is None:
            return None

        if self._backup_exchange is None:
            self._backup_exchange = self._backup_exchange_class()

        return self._backup_exchange

    @abstractmethod
    def fetch(self, symbol: str, start_timestamp: int) -> list:
        pass

    @abstractmethod
    def get_starting_time(self, symbol: str) -> int:
        pass



import math
import threading
import time
from datetime import timedelta
from typing import Dict, List, Any, Union

import arrow
import click
import pydash
from timeloop import Timeloop

import jesse.helpers as jh
from jesse.exceptions import CandleNotFoundInExchange
from jesse.models import Candle
from jesse.modes.import_candles_mode.drivers import drivers
from jesse.modes.import_candles_mode.drivers.interface import CandleExchange
from jesse.services.db import store_candles
from jesse.config import config
from jesse.services.failure import register_custom_exception_handler
from jesse.services.redis import sync_publish, process_status
from jesse.store import store
from jesse import exceptions
from jesse.services.progressbar import Progressbar


def run(exchange: str, symbol: str, start_date_str: str, skip_confirmation: bool = False, mode: str = 'candles') -> None:
    config['app']['trading_mode'] = mode

    # first, create and set session_id
    store.app.set_session_id()

    register_custom_exception_handler()

    # close database connection
    from jesse.services.db import database
    database.open_connection()

    # at every second, we check to see if it's time to execute stuff
    status_checker = Timeloop()

    @status_checker.job(interval=timedelta(seconds=1))
    def handle_time():
        if process_status() != 'started':
            raise exceptions.Termination

    status_checker.start()

    try:
        start_timestamp = jh.arrow_to_timestamp(arrow.get(start_date_str, 'YYYY-MM-DD'))
    except:
        raise ValueError('start_date must be a string representing a date before today. ex: 2020-01-17')

    # more start_date validations
    today = arrow.utcnow().floor('day').int_timestamp * 1000
    if start_timestamp == today:
        raise ValueError("Today's date is not accepted. start_date must be a string a representing date BEFORE today.")
    elif start_timestamp > today:
        raise ValueError("Future's date is not accepted. start_date must be a string a representing date BEFORE today.")

    # We just call this to throw a exception in case of a symbol without dash
    jh.quote_asset(symbol)

    click.clear()
    symbol = symbol.upper()

    until_date = arrow.utcnow().floor('day')
    start_date = arrow.get(start_timestamp / 1000)
    days_count = jh.date_diff_in_days(start_date, until_date)
    candles_count = days_count * 1440

    try:
        driver: CandleExchange = drivers[exchange]()
    except KeyError:
        raise ValueError(f'{exchange} is not a supported exchange')
    except TypeError:
        raise FileNotFoundError('You are missing the "plugins.py" file')

    loop_length = int(candles_count / driver.count) + 1
    # ask for confirmation
    if not skip_confirmation:
        click.confirm(
            f'Importing {days_count} days candles from "{exchange}" for "{symbol}". Duplicates will be skipped. All good?', abort=True, default=True)

    progressbar = Progressbar(loop_length)
    for i in range(candles_count):
        temp_start_timestamp = start_date.int_timestamp * 1000
        temp_end_timestamp = temp_start_timestamp + (driver.count - 1) * 60000

        # to make sure it won't try to import candles from the future! LOL
        if temp_start_timestamp > jh.now_to_timestamp():
            break

        # prevent duplicates calls to boost performance
        count = Candle.select().where(
            Candle.timestamp.between(temp_start_timestamp, temp_end_timestamp),
            Candle.symbol == symbol,
            Candle.exchange == exchange
        ).count()
        already_exists = count == driver.count

        if not already_exists:
            # it's today's candles if temp_end_timestamp < now
            if temp_end_timestamp > jh.now_to_timestamp():
                temp_end_timestamp = arrow.utcnow().floor('minute').int_timestamp * 1000 - 60000

            # fetch from market
            candles = driver.fetch(symbol, temp_start_timestamp)

            # check if candles have been returned and check those returned start with the right timestamp.
            # Sometimes exchanges just return the earliest possible candles if the start date doesn't exist.
            if not len(candles) or arrow.get(candles[0]['timestamp'] / 1000) > start_date:
                click.clear()
                first_existing_timestamp = driver.get_starting_time(symbol)

                # if driver can't provide accurate get_starting_time()
                if first_existing_timestamp is None:
                    raise CandleNotFoundInExchange(
                        f'No candles exists in the market for this day: {jh.timestamp_to_time(temp_start_timestamp)[:10]} \n'
                        'Try another start_date'
                    )

                # handle when there's missing candles during the period
                if temp_start_timestamp > first_existing_timestamp:
                    # see if there are candles for the same date for the backup exchange,
                    # if so, get those, if not, download from that exchange.
                    if driver.backup_exchange is not None:
                        candles = _get_candles_from_backup_exchange(
                            exchange, driver.backup_exchange, symbol, temp_start_timestamp, temp_end_timestamp
                        )

                else:
                    temp_start_time = jh.timestamp_to_time(temp_start_timestamp)[:10]
                    temp_existing_time = jh.timestamp_to_time(first_existing_timestamp)[:10]
                    sync_publish('alert', {
                        'message': f'No candle exists in the market for {temp_start_time}. So '
                                   f'Jesse started importing since the first existing date which is {temp_existing_time}',
                        'type': 'success'
                    })
                    run(exchange, symbol, jh.timestamp_to_time(first_existing_timestamp)[:10], True)
                    return

            # fill absent candles (if there's any)
            candles = _fill_absent_candles(candles, temp_start_timestamp, temp_end_timestamp)

            # store in the database
            if skip_confirmation:
                store_candles(candles)
            else:
                threading.Thread(target=store_candles, args=[candles]).start()

        # add as much as driver's count to the temp_start_time
        start_date = start_date.shift(minutes=driver.count)

        progressbar.update()
        sync_publish('progressbar', {
            'current': progressbar.current,
            'estimated_remaining_seconds': progressbar.estimated_remaining_seconds
        })

        # sleep so that the exchange won't get angry at us
        if not already_exists:
            time.sleep(driver.sleep_time)

    # stop the status_checker time loop
    status_checker.stop()

    sync_publish('alert', {
        'message': f'Successfully imported candles since {jh.timestamp_to_date(start_timestamp)} until today ({days_count} days). ',
        'type': 'success'
    })

    # if it is to skip, then it's being called from another process hence we should leave the database be
    if not skip_confirmation:
        # close database connection
        from jesse.services.db import database
        database.close_connection()


def _get_candles_from_backup_exchange(exchange: str, backup_driver: CandleExchange, symbol: str, start_timestamp: int,
                                      end_timestamp: int) -> List[Dict[str, Union[str, Any]]]:
    total_candles = []
    # try fetching from database first
    backup_candles = Candle.select(
        Candle.timestamp, Candle.open, Candle.close, Candle.high, Candle.low,
        Candle.volume
    ).where(
        Candle.timestamp.between(start_timestamp, end_timestamp),
        Candle.exchange == backup_driver.name,
        Candle.symbol == symbol).order_by(Candle.timestamp.asc()).tuples()
    already_exists = len(backup_candles) == (end_timestamp - start_timestamp) / 60_000 + 1
    if already_exists:
        # loop through them and set new ID and exchange
        for c in backup_candles:
            total_candles.append({
                'id': jh.generate_unique_id(),
                'symbol': symbol,
                'exchange': exchange,
                'timestamp': c[0],
                'open': c[1],
                'close': c[2],
                'high': c[3],
                'low': c[4],
                'volume': c[5]
            })

        return total_candles

    # try fetching from market now
    days_count = jh.date_diff_in_days(jh.timestamp_to_arrow(start_timestamp), jh.timestamp_to_arrow(end_timestamp))
    # make sure it's rounded up so that we import maybe more candles, but not less
    days_count = max(days_count, 1)
    if type(days_count) is float and not days_count.is_integer():
        days_count = math.ceil(days_count)
    candles_count = days_count * 1440
    start_date = jh.timestamp_to_arrow(start_timestamp).floor('day')
    for _ in range(candles_count):
        temp_start_timestamp = start_date.int_timestamp * 1000
        temp_end_timestamp = temp_start_timestamp + (backup_driver.count - 1) * 60000

        # to make sure it won't try to import candles from the future! LOL
        if temp_start_timestamp > jh.now_to_timestamp():
            break

        # prevent duplicates
        count = Candle.select().where(
            Candle.timestamp.between(temp_start_timestamp, temp_end_timestamp),
            Candle.symbol == symbol,
            Candle.exchange == backup_driver.name
        ).count()
        already_exists = count == backup_driver.count

        if not already_exists:
            # it's today's candles if temp_end_timestamp < now
            if temp_end_timestamp > jh.now_to_timestamp():
                temp_end_timestamp = arrow.utcnow().floor('minute').int_timestamp * 1000 - 60000

            # fetch from market
            candles = backup_driver.fetch(symbol, temp_start_timestamp)

            if not len(candles):
                raise CandleNotFoundInExchange(
                    f'No candles exists in the market for this day: {jh.timestamp_to_time(temp_start_timestamp)[:10]} \n'
                    'Try another start_date'
                )

            # fill absent candles (if there's any)
            candles = _fill_absent_candles(candles, temp_start_timestamp, temp_end_timestamp)

            # store in the database
            store_candles(candles)

        # add as much as driver's count to the temp_start_time
        start_date = start_date.shift(minutes=backup_driver.count)

        # sleep so that the exchange won't get angry at us
        if not already_exists:
            time.sleep(backup_driver.sleep_time)

    # now try fetching from database again. Why? because we might have fetched more
    # than what's needed, but we only want as much was requested. Don't worry, the next
    # request will probably fetch from database and there won't be any waste!
    backup_candles = Candle.select(
        Candle.timestamp, Candle.open, Candle.close, Candle.high, Candle.low,
        Candle.volume
    ).where(
        Candle.timestamp.between(start_timestamp, end_timestamp),
        Candle.exchange == backup_driver.name,
        Candle.symbol == symbol).order_by(Candle.timestamp.asc()).tuples()
    already_exists = len(backup_candles) == (end_timestamp - start_timestamp) / 60_000 + 1
    if already_exists:
        # loop through them and set new ID and exchange
        for c in backup_candles:
            total_candles.append({
                'id': jh.generate_unique_id(),
                'symbol': symbol,
                'exchange': exchange,
                'timestamp': c[0],
                'open': c[1],
                'close': c[2],
                'high': c[3],
                'low': c[4],
                'volume': c[5]
            })

        return total_candles


def _fill_absent_candles(temp_candles: List[Dict[str, Union[str, Any]]], start_timestamp: int, end_timestamp: int) -> List[Dict[str, Union[str, Any]]]:
    if not temp_candles:
        raise CandleNotFoundInExchange(
            f'No candles exists in the market for this day: {jh.timestamp_to_time(start_timestamp)[:10]} \n'
            'Try another start_date'
        )

    symbol = temp_candles[0]['symbol']
    exchange = temp_candles[0]['exchange']
    candles = []
    first_candle = temp_candles[0]
    started = False
    loop_length = ((end_timestamp - start_timestamp) / 60000) + 1

    for _ in range(int(loop_length)):
        candle_for_timestamp = pydash.find(
            temp_candles, lambda c: c['timestamp'] == start_timestamp)

        if candle_for_timestamp is None:
            if started:
                last_close = candles[-1]['close']
                candles.append({
                    'id': jh.generate_unique_id(),
                    'symbol': symbol,
                    'exchange': exchange,
                    'timestamp': start_timestamp,
                    'open': last_close,
                    'high': last_close,
                    'low': last_close,
                    'close': last_close,
                    'volume': 0
                })
            else:
                candles.append({
                    'id': jh.generate_unique_id(),
                    'symbol': symbol,
                    'exchange': exchange,
                    'timestamp': start_timestamp,
                    'open': first_candle['open'],
                    'high': first_candle['open'],
                    'low': first_candle['open'],
                    'close': first_candle['open'],
                    'volume': 0
                })
        # candle is present
        else:
            started = True
            candles.append(candle_for_timestamp)

        start_timestamp += 60000
    return candles





from typing import Union

import numpy as np

try:
    from numba import njit
except ImportError:
    njit = lambda a: a

from jesse.helpers import get_candle_source, slice_candles


def nma(candles: np.ndarray, period: int = 40, source_type: str = "close", sequential: bool = False) -> Union[
    float, np.ndarray]:
    """
    Natural Moving Average

    :param candles: np.ndarray
    :param period: int - default: 40
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """

    # Accept normal array too.
    if len(candles.shape) == 1:
        source = candles
    else:
        candles = slice_candles(candles, sequential)
        source = get_candle_source(candles, source_type=source_type)

    res = nma_fast(source, period)

    return res if sequential else res[-1]


@njit
def nma_fast(source, period):
    ln = np.log(source) * 1000
    newseries = np.full_like(source, np.nan)

    for j in range(period + 1, source.shape[0]):
        num = 0.0
        denom = 0.0
        for i in range(period):
            oi = np.abs(ln[j - i] - ln[j - i - 1])
            num += oi * (np.sqrt(i + 1) - np.sqrt(i))
            denom += oi

        ratio = num / denom if denom != 0 else 0
        newseries[j] = (source[j - i] * ratio) + (source[j - i - 1] * (1 - ratio))

    return newseries



import os
import shutil
from starlette.responses import JSONResponse


def generate(name: str) -> JSONResponse:
    path = f'strategies/{name}'

    # validation for name duplication
    exists = os.path.isdir(path)
    if exists:
        return JSONResponse({
            'status': 'error',
            'message': f'Strategy "{name}" already exists.'
        })

    # generate from ExampleStrategy
    dirname, filename = os.path.split(os.path.abspath(__file__))

    shutil.copytree(f'{dirname}/ExampleStrategy', path)

    # replace 'ExampleStrategy' with the name of the new strategy
    with open(f"{path}/__init__.py", "rt") as fin:
        data = fin.read()
        data = data.replace('ExampleStrategy', name)
    with open(f"{path}/__init__.py", "wt") as fin:
        fin.write(data)
    # return the location of generated strategy directory
    return JSONResponse({
        'status': 'success',
        'message': path
    })



from collections import namedtuple

import numpy as np
import tulipy as ti

from jesse.helpers import get_candle_source, same_length, slice_candles

MSW = namedtuple('MSW', ['sine', 'lead'])


def msw(candles: np.ndarray, period: int = 5, source_type: str = "close", sequential: bool = False) -> MSW:
    """
    MSW - Mesa Sine Wave

    :param candles: np.ndarray
    :param period: int - default: 5
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: MSW(sine, lead)
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)
    msw_sine, msw_lead = ti.msw(np.ascontiguousarray(source), period=period)

    s = same_length(candles, msw_sine)
    l = same_length(candles, msw_lead)

    if sequential:
        return MSW(s, l)
    else:
        return MSW(s[-1], l[-1])



from collections import namedtuple

import numpy as np
try:
    from numba import njit
except ImportError:
    njit = lambda a : a

from .high_pass import high_pass_fast

from jesse.helpers import get_candle_source, slice_candles

BandPass = namedtuple('BandPass', ['bp', 'bp_normalized', 'signal', 'trigger'])


def bandpass(candles: np.ndarray, period: int = 20, bandwidth: float = 0.3,  source_type: str = "close",  sequential: bool = False) -> BandPass:
    """
    BandPass Filter

    :param candles: np.ndarray
    :param period: int - default: 20
    :param bandwidth: float - default: 0.3
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: BandPass(bp, bp_normalized, signal, trigger)
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)


    hp = high_pass_fast(source, 4 * period / bandwidth)

    beta = np.cos(2 * np.pi / period)
    gamma = np.cos(2 * np.pi * bandwidth / period)
    alpha = 1 / gamma - np.sqrt(1 / gamma ** 2 - 1)

    bp, peak = bp_fast(source, hp, alpha, beta)

    bp_normalized = bp / peak

    trigger = high_pass_fast(bp_normalized, period / bandwidth / 1.5)
    signal = (bp_normalized < trigger) * 1 - (trigger < bp_normalized) * 1

    if sequential:
        return BandPass(bp, bp_normalized, signal, trigger)
    else:
        return BandPass(bp[-1], bp_normalized[-1], signal[-1], trigger[-1])


@njit
def bp_fast(source, hp, alpha, beta):  # Function is compiled to machine code when called the first time

    bp = np.copy(hp)
    for i in range(2, source.shape[0]):
      bp[i] = 0.5 * (1 - alpha) * hp[i] - (1 - alpha) * 0.5 * hp[i - 2]  + beta * (1 + alpha) * bp[i - 1] - alpha * bp[i - 2]

    # fast attack-slow decay AGC
    K = 0.991
    peak = np.copy(bp)
    for i in range(source.shape[0]):
      if i > 0:
        peak[i] = peak[i - 1] * K
      if np.abs(bp[i]) > peak[i]:
        peak[i] = np.abs(bp[i])

    return bp, peak



import jesse.helpers as jh
from jesse.config import config
from jesse.exceptions import InvalidConfig
from jesse.models import SpotExchange, FuturesExchange
import pydash


class ExchangesState:
    def __init__(self) -> None:
        self.storage = {}

        for name in config['app']['considering_exchanges']:
            starting_assets = config['env']['exchanges'][name]['assets']
            fee = config['env']['exchanges'][name]['fee']
            exchange_type = config['env']['exchanges'][name]['type']

            if exchange_type == 'spot':
                self.storage[name] = SpotExchange(name, starting_assets, fee)
            elif exchange_type == 'futures':
                settlement_currency = jh.get_config(f'env.exchanges.{name}.settlement_currency')
                # dirty fix to get the settlement_currency right for none-USDT pairs
                settlement_asset_dict = pydash.find(starting_assets, lambda asset: asset['asset'] == settlement_currency)
                if settlement_asset_dict is None:
                    starting_assets[0]['asset'] = settlement_currency
                self.storage[name] = FuturesExchange(
                    name, starting_assets, fee,
                    settlement_currency=settlement_currency,
                    futures_leverage_mode=jh.get_config(f'env.exchanges.{name}.futures_leverage_mode'),
                    futures_leverage=jh.get_config(f'env.exchanges.{name}.futures_leverage'),
                )
            else:
                raise InvalidConfig(
                    'Value for exchange type in your config file in not valid. Supported values are "spot" and "futures"')



import numpy as np

from jesse.strategies import Strategy


class TestLiquidationInCrossModeForShortTrade(Strategy):
    def on_open_position(self, order):
        # print(self.position.liquidation_price)
        # assert self.position.liquidation_price == 22.09
        assert np.isnan(self.position.liquidation_price)

    def before(self):
        if self.index == 0:
            assert self.balance == 10000
            assert self.leverage == 10
            assert self.available_margin == 10 * 10000
            assert self.position.mode == 'cross'

        # # the liquidation price is at $14.94 so at $15:
        # if self.price == 15:
        #     # assert that we are liquidated by this point
        #     print(self.balance)
        #     print(self.available_margin)
        #     assert self.is_close
        #     assert self.balance == 0
        #     assert self.available_margin == 0

    def should_long(self) -> bool:
        return False

    def should_short(self) -> bool:
        return self.price == 20

    def go_long(self):
        pass

    def go_short(self):
        qty = self.available_margin / self.price
        print('qty', qty)
        self.sell = qty, self.price

    def should_cancel(self):
        return False



class CompletedTrades:
    def __init__(self) -> None:
        self.trades = []

    def add_trade(self, trade) -> None:
        self.trades.append(trade)

    @property
    def count(self) -> int:
        return len(self.trades)



import pytest

from jesse.exceptions import NegativeBalance, InsufficientMargin
from .utils import single_route_backtest


def test_negative_balance_validation_for_spot_market():
    """
    The initial account balance for USDT is 10_000. So trying to spend 10_001
    should throw NegativeBalance exception
    """
    with pytest.raises(NegativeBalance):
        single_route_backtest('TestSpotNegativeBalance', is_futures_trading=False)


def test_negative_balance_validation_for_futures_market():
    """
    The initial account balance for USDT is 10_000. So trying to spend 10_001
    should throw InsufficientMargin exception
    """
    # (MARKET order)
    with pytest.raises(InsufficientMargin):
        single_route_backtest('TestInsufficientMargin1', is_futures_trading=True, leverage=1)
    # but with more leverage, it should work
    single_route_backtest('TestInsufficientMargin1', is_futures_trading=True, leverage=2)

    # (LIMIT order)
    with pytest.raises(InsufficientMargin):
        single_route_backtest('TestInsufficientMargin2', is_futures_trading=True, leverage=1)
    # but with more leverage, it should work
    single_route_backtest('TestInsufficientMargin2', is_futures_trading=True, leverage=2)

    # short-version (STOP order)
    with pytest.raises(InsufficientMargin):
        single_route_backtest('TestInsufficientMargin3', is_futures_trading=True, leverage=1)
    # but with more leverage, it should work
    single_route_backtest('TestInsufficientMargin3', is_futures_trading=True, leverage=2)


def test_wallet_balance_and_available_margin_for_futures_market():
    """
    Works the same way as Binance Futures'es wallet balance does.
    It is only changed at commission charges and after realized PNL is added.
    """
    single_route_backtest('TestWalletBalance', is_futures_trading=True, leverage=2)


def test_available_margin_for_futures_market():
    """
    Make sure that the reduce only orders are not affecting the available_margin causing it to go negative
    """
    single_route_backtest('TestFuturesExchangeAvailableMargin', is_futures_trading=True, leverage=2)


# TODO: test_wallet_balance_and_available_margin_for_spot_market



from jesse.strategies import Strategy


class TestMarketOrderForLowPriceDifference(Strategy):
    def on_open_position(self, order):
        assert order.type == 'MARKET'

    def on_close_position(self, order) -> None:
        assert order.type == 'MARKET'

    def should_long(self) -> bool:
        return self.index == 0

    def should_short(self) -> bool:
        return False

    def go_long(self):
        # current-price: 1
        self.buy = 1, 1.00001

    def update_position(self) -> None:
        # submit a take-profit order that has a low difference with the current price
        self.take_profit = self.position.qty, self.price + 0.00001

    def go_short(self):
        pass

    def should_cancel(self):
        return False



from jesse import utils
from jesse.strategies import Strategy


class TestReduceOnlyMarketOrders(Strategy):
    def should_long(self) -> bool:
        return self.price == 10

    def should_short(self) -> bool:
        return False

    def go_long(self):
        entry = self.price
        qty = utils.size_to_qty(self.capital, entry, fee_rate=self.fee_rate)
        self.buy = qty, entry

    def go_short(self):
        pass

    def should_cancel(self):
        return False

    def update_position(self):
        if self.price == 20:
            self.liquidate()



import jesse.helpers as jh
from jesse.config import config
from jesse.exceptions import InvalidRoutes
from jesse.routes import router
from .state_app import AppState
from .state_candles import CandlesState
from .state_completed_trades import CompletedTrades
from .state_exchanges import ExchangesState
from .state_logs import LogsState
from .state_orderbook import OrderbookState
from .state_orders import OrdersState
from .state_positions import PositionsState
from .state_tickers import TickersState
from .state_trades import TradesState


def install_routes() -> None:
    considering_candles = set()

    # when importing market data, considering_candles is all we need
    if jh.is_collecting_data():
        for r in router.market_data:
            considering_candles.add((r.exchange, r.symbol))

        config['app']['considering_candles'] = tuple(considering_candles)
        return

    # validate routes for duplicates:
    # each exchange-symbol pair can be traded only once.
    for r in router.routes:
        considering_candles.add((r.exchange, r.symbol))

        exchange = r.exchange
        symbol = r.symbol
        count = sum(
            ro.exchange == exchange and ro.symbol == symbol
            for ro in router.routes
        )

        if count != 1:
            raise InvalidRoutes(
                'each exchange-symbol pair can be traded only once. \nMore info: https://docs.jesse.trade/docs/routes.html#trading-multiple-routes')

    # check to make sure if trading more than one route, they all have the same quote
    # currency because otherwise we cannot calculate the correct performance metrics
    first_routes_quote = jh.quote_asset(router.routes[0].symbol)
    for r in router.routes:
        if jh.quote_asset(r.symbol) != first_routes_quote:
            raise InvalidRoutes('All trading routes must have the same quote asset.')

    trading_exchanges = set()
    trading_timeframes = set()
    trading_symbols = set()

    for r in router.routes:
        trading_exchanges.add(r.exchange)
        trading_timeframes.add(r.timeframe)
        trading_symbols.add(r.symbol)

    considering_exchanges = trading_exchanges.copy()
    considering_timeframes = trading_timeframes.copy()
    considering_symbols = trading_symbols.copy()

    for e in router.extra_candles:
        considering_candles.add((e['exchange'], e['symbol']))
        considering_exchanges.add(e['exchange'])
        considering_symbols.add(e['symbol'])
        considering_timeframes.add(e['timeframe'])

    # 1m must be present at all times
    considering_timeframes.add('1m')

    config['app']['considering_candles'] = tuple(considering_candles)
    config['app']['considering_exchanges'] = tuple(considering_exchanges)

    config['app']['considering_symbols'] = tuple(considering_symbols)
    config['app']['considering_timeframes'] = tuple(considering_timeframes)
    config['app']['trading_exchanges'] = tuple(trading_exchanges)
    config['app']['trading_symbols'] = tuple(trading_symbols)
    config['app']['trading_timeframes'] = tuple(trading_timeframes)


class StoreClass:
    app = AppState()
    orders = OrdersState()
    completed_trades = CompletedTrades()
    logs = LogsState()
    exchanges = ExchangesState()
    candles = CandlesState()
    positions = PositionsState()
    tickers = TickersState()
    trades = TradesState()
    orderbooks = OrderbookState()

    def __init__(self) -> None:
        self.vars = {}

    def reset(self, force_install_routes: bool = False) -> None:
        """
        Resets all the states within the store
        
        Keyword Arguments:
            force_install_routes {bool} -- used for unit_testing (default: {False})
        """
        if not jh.is_unit_testing() or force_install_routes:
            install_routes()

        self.app = AppState()
        self.orders = OrdersState()
        self.completed_trades = CompletedTrades()
        self.logs = LogsState()
        self.exchanges = ExchangesState()
        self.candles = CandlesState()
        self.positions = PositionsState()
        self.tickers = TickersState()
        self.trades = TradesState()
        self.orderbooks = OrderbookState()


# if not jh.is_unit_testing():
#     install_routes()

store = StoreClass()
# store.reset()



class Route:
    def __init__(self, exchange: str, symbol: str, timeframe: str = None, strategy_name: str = None,
                 dna: str = None) -> None:
        # replace PERP with USD in FTX routes
        if exchange.startswith('FTX') and symbol.upper().endswith('PERP'):
            symbol = symbol.replace('PERP', 'USD')

        self.exchange = exchange
        self.symbol = symbol
        self.timeframe = timeframe
        self.strategy_name = strategy_name
        self.strategy = None
        self.dna = dna



from typing import Union

import numpy as np
import talib

from jesse.helpers import get_candle_source
from jesse.helpers import slice_candles


def ui(candles: np.ndarray, period: int = 14, scalar: float = 100, source_type: str = "close",  sequential: bool = False) -> Union[float, np.ndarray]:
    """
    Ulcer Index (UI)

    :param candles: np.ndarray
    :param period: int - default: 14
    :param scalar: float - default: 100
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)
    source = get_candle_source(candles, source_type=source_type)

    highest_close = talib.MAX(source, period)
    downside = scalar * (source - highest_close)
    downside /= highest_close
    d2 = downside * downside

    res = np.sqrt(talib.SUM(d2, period) / period)

    return res if sequential else res[-1]



from jesse.strategies import Strategy


# test_on_route_canceled part 2 - ETH-USD
class Test28(Strategy):
    def should_long(self):
        return self.price == 10

    def should_short(self):
        return False

    def go_long(self):
        # because we know the price is going up, this order will never get filled
        self.buy = 1, self.price - 10

    def go_short(self):
        pass

    def should_cancel(self):
        return self.price == 20



from collections import namedtuple

import numpy as np

from jesse.helpers import slice_candles

PIVOT = namedtuple('PIVOT', ['r4', 'r3', 'r2', 'r1', 'pp', 's1', 's2', 's3', 's4'])


def pivot(candles: np.ndarray, mode: int = 0, sequential: bool = False) -> PIVOT:
    """
    Pivot Points

    :param candles: np.ndarray
    :param mode: int - default = 0
    :param sequential: bool - default: False

    :return: PIVOT(r4, r3, r2, r1, pp, s1, s2, s3, s4)
    """
    candles = slice_candles(candles, sequential)

    high = candles[:, 3]
    low = candles[:, 4]
    close = candles[:, 2]
    open = candles[:, 1]

    s1 = np.full_like(close, np.nan)
    s2 = np.copy(s1)
    s3 = np.copy(s1)
    s4 = np.copy(s1)
    p = np.copy(s1)
    r1 = np.copy(s1)
    r2 = np.copy(s1)
    r3 = np.copy(s1)
    r4 = np.copy(s1)

    # Standard Pivot Points / Floor Pivot Points
    if mode == 0:
        p = (high + low + close) / 3
        s1 = (2 * p) - high
        s2 = p - (high - low)
        r1 = (2 * p) - low
        r2 = p + (high - low)
    # Fibonacci Pivot Points
    elif mode == 1:
        p = (high + low + close) / 3
        s1 = p - 0.382 * (high - low)
        s2 = p - 0.618 * (high - low)
        s3 = p - 1 * (high - low)
        r1 = p + 0.382 * (high - low)
        r2 = p + 0.618 * (high - low)
        r3 = p + 1 * (high - low)
    # Demark Pivot Points
    elif mode == 2:
        p = np.where(close < open, (high + 2 * low + close) / 4, np.where(close > open, (2 * high + low + close) / 4,
                                                                          np.where(close == open,
                                                                                   (high + low + 2 * close) / 4,
                                                                                   np.nan)))
        s1 = np.where(close < open, (high + 2 * low + close) / 2 - high,
                      np.where(close > open, (2 * high + low + close) / 2 - high,
                               np.where(close == open, (high + low + 2 * close) / 2 - high, np.nan)))
        r1 = np.where(close < open, (high + 2 * low + close) / 2 - low,
                      np.where(close > open, (2 * high + low + close) / 2 - low,
                               np.where(close == open, (high + low + 2 * close) / 2 - low, np.nan)))
    elif mode == 3:
        # Camarilla Pivot Points
        p = (high + low + close) / 3
        r4 = (0.55 * (high - low)) + close
        r3 = (0.275 * (high - low)) + close
        r2 = (0.183 * (high - low)) + close
        r1 = (0.0916 * (high - low)) + close
        s1 = close - (0.0916 * (high - low))
        s2 = close - (0.183 * (high - low))
        s3 = close - (0.275 * (high - low))
        s4 = close - (0.55 * (high - low))
    # Woodie's Pivot Points
    elif mode == 4:
        p = (high + low + (2 * open)) / 4
        r3 = high + 2 * (p - low)
        r4 = r3 + (high - low)
        r2 = p + (high - low)
        r1 = (2 * p) - low
        s1 = (2 * p) - high
        s2 = p - (high - low)
        s3 = low - 2 * (high - p)
        s4 = s3 - (high - low)

    if sequential:
        return PIVOT(r4, r3, r2, r1, p, s1, s2, s3, s4)
    else:
        return PIVOT(r4[-1], r3[-1], r2[-1], r1[-1], p[-1], s1[-1], s2[-1], s3[-1], s4[-1])



import os
from hashlib import sha256
import requests
from fastapi.responses import JSONResponse
from jesse.services.env import ENV_VALUES
import jesse.helpers as jh


def password_to_token(password: str) -> JSONResponse:
    if password != ENV_VALUES['PASSWORD']:
        return unauthorized_response()

    auth_token = sha256(password.encode('utf-8')).hexdigest()

    return JSONResponse({
        'auth_token': auth_token,
    }, status_code=200)


def is_valid_token(auth_token: str) -> bool:
    hashed_local_pass = sha256(ENV_VALUES['PASSWORD'].encode('utf-8')).hexdigest()
    return auth_token == hashed_local_pass


def unauthorized_response() -> JSONResponse:
    return JSONResponse({
        'message': "Invalid password",
    }, status_code=401)


def get_access_token():
    from jesse.services.env import ENV_VALUES

    if 'LICENSE_API_TOKEN' not in ENV_VALUES:
        return None
    if not ENV_VALUES['LICENSE_API_TOKEN']:
        return None

    return ENV_VALUES['LICENSE_API_TOKEN']


def login_to_jesse_trade(email: str, password: str) -> JSONResponse:
    response = requests.post('https://jesse.trade/api/auth', json={
        'email': email,
        'password': password,
        'device_name': jh.computer_name()
    })

    if response.status_code != 200:
        return JSONResponse({
            'status': 'error',
            'message': f"[{response.status_code}]: {response.json()['message']}\nPlease contact us at support@jesse.trade if this is unexpected.",
        }, status_code=200)

    access_token = response.json()['token']

    os.makedirs('./storage/temp', exist_ok=True)
    with open('storage/temp/access_token', 'w') as f:
        f.write(access_token)

    return JSONResponse({
        'status': 'success',
        'message': 'Successfully logged in to Jesse.Trade',
    }, status_code=200)


def logout_from_jesse_trade() -> JSONResponse:
    os.remove('storage/temp/access_token')

    return JSONResponse({
        'message': 'Successfully logged out from Jesse.Trade and removed access_token file'
    })



from typing import List, ValuesView, Optional, Any


def get_current_price(exchange: str, symbol: str) -> float:
    p = get_position(exchange, symbol)
    return p.current_price


def get_position(exchange: str, symbol: str) -> Any:
    from jesse.store import store
    key = f'{exchange}-{symbol}'
    return store.positions.storage.get(key, None)


def get_orders(exchange: str, symbol: str) -> List[Any]:
    from jesse.store import store
    return store.orders.get_orders(exchange, symbol)


def get_time() -> int:
    from jesse.store import store
    return store.app.time


def get_exchange(name: str) -> Any:
    from jesse.store import store
    return store.exchanges.storage.get(name, None)


def get_all_exchanges() -> ValuesView[Any]:
    from jesse.store import store
    return store.exchanges.storage.values()


def get_strategy(exchange: str, symbol: str) -> Any:
    from jesse.routes import router
    r = next(r for r in router.routes
             if r.exchange == exchange and r.symbol == symbol)
    return r.strategy


def get_route(exchange: str, symbol: str) -> Optional[Any]:
    from jesse.routes import router
    return next(
        (
            r
            for r in router.routes
            if r.exchange == exchange and r.symbol == symbol
        ),
        None,
    )


def get_all_trading_routes() -> List[Any]:
    from jesse.routes import router
    return router.routes



import numpy as np

import jesse.helpers as jh
from jesse.config import config
from jesse.libs import DynamicNumpyArray
from jesse.models import store_orderbook_into_db


class OrderbookState:
    def __init__(self) -> None:
        self.storage = {}
        self.temp_storage = {}

    def init_storage(self) -> None:
        for c in config['app']['considering_candles']:
            key = jh.key(c[0], c[1])
            self.temp_storage[key] = {
                'last_updated_timestamp': None,
                'asks': [],
                'bids': []
            }
            self.storage[key] = DynamicNumpyArray((60, 2, 50, 2), drop_at=60)

    def format_orderbook(self, exchange: str, symbol: str) -> np.ndarray:
        key = jh.key(exchange, symbol)

        # trim prices
        asks = _trim_orderbook_list(self.temp_storage[key]['asks'], ascending=True)
        bids = _trim_orderbook_list(self.temp_storage[key]['bids'], ascending=False)

        # fill empty values with NaN
        asks = _fix_array_len(np.array(asks), 50)
        bids = _fix_array_len(np.array(bids), 50)

        return np.array([
            asks, bids
        ])

    def add_orderbook(self, exchange: str, symbol: str, asks: list, bids: list) -> None:
        key = jh.key(exchange, symbol)
        self.temp_storage[key]['asks'] = asks
        self.temp_storage[key]['bids'] = bids

        # generate new numpy formatted orderbook if it is
        # either the first time, or that it has passed
        # 1000 milliseconds since the last time
        if self.temp_storage[key]['last_updated_timestamp'] is None or jh.now_to_timestamp() - self.temp_storage[key][
            'last_updated_timestamp'] >= 1000:
            self.temp_storage[key]['last_updated_timestamp'] = jh.now_to_timestamp()

            formatted_orderbook = self.format_orderbook(exchange, symbol)

            if jh.is_collecting_data():
                store_orderbook_into_db(exchange, symbol, formatted_orderbook)
            else:
                self.storage[key].append(formatted_orderbook)

    def get_current_orderbook(self, exchange: str, symbol: str) -> np.ndarray:
        key = jh.key(exchange, symbol)
        return self.storage[key][-1]

    def get_current_asks(self, exchange: str, symbol: str) -> np.ndarray:
        key = jh.key(exchange, symbol)
        return self.storage[key][-1][0]

    def get_best_ask(self, exchange: str, symbol: str) -> np.ndarray:
        key = jh.key(exchange, symbol)
        return self.storage[key][-1][0][0]

    def get_current_bids(self, exchange: str, symbol: str) -> np.ndarray:
        key = jh.key(exchange, symbol)
        return self.storage[key][-1][1]

    def get_best_bid(self, exchange: str, symbol: str) -> np.ndarray:
        key = jh.key(exchange, symbol)
        return self.storage[key][-1][1][0]

    def get_orderbooks(self, exchange: str, symbol: str) -> np.ndarray:
        key = jh.key(exchange, symbol)
        return self.storage[key][:]


def _trim_orderbook_list(arr: list, ascending: bool, limit_len: int = 50) -> list:
    """trims prices up to 4 digits precision"""
    first_price = arr[0][0]
    if first_price < 0.1:
        unit = 1e-5
    elif first_price < 1:
        unit = 1e-4
    elif first_price < 10:
        unit = 1e-3
    elif first_price < 100:
        unit = 1e-2
    elif first_price < 1000:
        unit = 1e-1
    elif first_price < 10000:
        unit = 1
    else:
        unit = 10

    trimmed_price = jh.orderbook_trim_price(first_price, ascending, unit)
    temp_qty = 0
    trimmed_arr = []
    for a in arr:
        if len(trimmed_arr) == limit_len:
            break

        if (ascending and a[0] > trimmed_price) or (not ascending and a[0] < trimmed_price):
            # add previous record
            trimmed_arr.append([
                trimmed_price, temp_qty
            ])
            # update temp values
            temp_qty = a[1]
            trimmed_price = jh.orderbook_trim_price(a[0], ascending, unit)
        else:
            temp_qty += a[1]

    return trimmed_arr


def _fix_array_len(arr: np.ndarray, target_len: int) -> np.ndarray:
    """make sure bids and asks have the same length"""
    missing_len = target_len - len(arr)

    if missing_len < 0:
        raise ValueError(f"len cannot be smaller than array's length. {target_len} sent, while array has {len(arr)} items")

    if not missing_len:
        return arr

    return np.array([*arr, *np.full((missing_len, *arr.shape[1:]), np.nan)])



import requests

import jesse.helpers as jh
from jesse import exceptions
from jesse.modes.import_candles_mode.drivers.interface import CandleExchange


class Bitfinex(CandleExchange):
    def __init__(self) -> None:
        # import here instead of the top of the file to prevent possible the circular imports issue
        from jesse.modes.import_candles_mode.drivers.coinbase import Coinbase

        super().__init__(
            name='Bitfinex',
            count=1440,
            rate_limit_per_second=1,
            backup_exchange_class=Coinbase
        )

        self.endpoint = 'https://api-pub.bitfinex.com/v2/candles'

    def get_starting_time(self, symbol: str) -> int:
        dashless_symbol = jh.dashless_symbol(symbol)

        # hard-code few common symbols
        if symbol == 'BTC-USD':
            return jh.date_to_timestamp('2015-08-01')
        elif symbol == 'ETH-USD':
            return jh.date_to_timestamp('2016-01-01')

        payload = {
            'sort': 1,
            'limit': 5000,
        }

        response = requests.get(f"{self.endpoint}/trade:1D:t{dashless_symbol}/hist", params=payload)

        if response.status_code != 200:
            raise Exception(response.content)

        data = response.json()

        # wrong symbol entered
        if not len(data):
            raise exceptions.SymbolNotFound(
                f"No candle exists for {symbol} in Bitfinex. You're probably misspelling the symbol name."
            )

        # since the first timestamp doesn't include all the 1m
        # candles, let's start since the second day then
        first_timestamp = int(data[0][0])
        return first_timestamp + 60_000 * 1440

    def fetch(self, symbol: str, start_timestamp: int) -> list:
        # since Bitfinex API skips candles with "volume=0", we have to send end_timestamp
        # instead of limit. Therefore, we use limit number to calculate the end_timestamp
        end_timestamp = start_timestamp + (self.count - 1) * 60000

        payload = {
            'start': start_timestamp,
            'end': end_timestamp,
            'limit': self.count,
            'sort': 1
        }

        dashless_symbol = jh.dashless_symbol(symbol)

        response = requests.get(
            f"{self.endpoint}/trade:1m:t{dashless_symbol}/hist",
            params=payload
        )

        data = response.json()
        return [{
                'id': jh.generate_unique_id(),
                'symbol': symbol,
                'exchange': self.name,
                'timestamp': d[0],
                'open': d[1],
                'close': d[2],
                'high': d[3],
                'low': d[4],
                'volume': d[5]
            } for d in data]



import os
from datetime import datetime, timedelta

import arrow
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from pandas.plotting import register_matplotlib_converters

import jesse.helpers as jh
from jesse.config import config
from jesse.routes import router
from jesse.store import store


def equity_curve() -> list:
    start_date = datetime.fromtimestamp(store.app.starting_time / 1000)
    date_list = [start_date + timedelta(days=x) for x in range(len(store.app.daily_balance))]
    daily_balance = store.app.daily_balance

    return [{
        'timestamp': date.timestamp(),
        'balance': balance
    } for date, balance in zip(date_list, daily_balance)]


def portfolio_vs_asset_returns(study_name: str = None) -> str:
    if jh.is_unit_testing():
        return 'charts'

    register_matplotlib_converters()
    trades = store.completed_trades.trades
    # create a plot figure
    plt.figure(figsize=(26, 16))

    # daily balance
    plt.subplot(2, 1, 1)
    start_date = datetime.fromtimestamp(store.app.starting_time / 1000)
    date_list = [start_date + timedelta(days=x) for x in range(len(store.app.daily_balance))]
    plt.xlabel('date')
    plt.ylabel('balance')
    if study_name:
        plt.title(f'Portfolio Daily Return - {study_name}')
    else:
        plt.title('Portfolio Daily Return')
    plt.plot(date_list, store.app.daily_balance)

    # price change%
    plt.subplot(2, 1, 2)
    price_dict = {}
    for r in router.routes:
        key = jh.key(r.exchange, r.symbol)
        price_dict[key] = {
            'indexes': {},
            'prices': []
        }
        dates = []
        prices = []
        candles = store.candles.get_candles(r.exchange, r.symbol, '1m')
        max_timeframe = jh.max_timeframe(config['app']['considering_timeframes'])
        pre_candles_count = jh.timeframe_to_one_minutes(max_timeframe) * jh.get_config(
            'env.data.warmup_candles_num', 210
        )
        for i, c in enumerate(candles):
            # do not plot prices for required_initial_candles period
            if i < pre_candles_count:
                continue

            dates.append(datetime.fromtimestamp(c[0] / 1000))
            prices.append(c[2])
            # save index of the price instead of the actual price
            price_dict[key]['indexes'][str(int(c[0]))] = len(prices) - 1

        # price => %returns
        price_returns = pd.Series(prices).pct_change(1) * 100
        cumsum_returns = np.cumsum(price_returns)
        if len(router.routes) == 1:
            plt.plot(dates, cumsum_returns, label=r.symbol, c='grey')
        else:
            plt.plot(dates, cumsum_returns, label=r.symbol)
        price_dict[key]['prices'] = cumsum_returns

    # buy and sell plots
    buy_x = []
    buy_y = []
    sell_x = []
    sell_y = []
    for index, t in enumerate(trades):
        key = jh.key(t.exchange, t.symbol)

        # dirty fix for an issue with last trade being an open trade at the end of backtest
        if index == len(trades) - 1 and store.app.total_open_trades > 0:
            continue

        if t.type == 'long':
            # Buy
            if str(int(t.opened_at)) in price_dict[key]['indexes']:
                # add price change%
                buy_y.append(
                    price_dict[key]['prices'][price_dict[key]['indexes'][str(int(t.opened_at))]]
                )
                # add datetime
                buy_x.append(datetime.fromtimestamp(t.opened_at / 1000))

            # Sell: only generate data point if this trade wasn't after the last candle (open position at end)
            if str(int(t.closed_at)) in price_dict[key]['indexes']:
                # add price change%
                sell_y.append(
                    price_dict[key]['prices'][price_dict[key]['indexes'][str(int(t.closed_at))]]
                )
                # add datetime
                sell_x.append(datetime.fromtimestamp(t.closed_at / 1000))

        elif t.type == 'short':
            # Buy: only generate data point if this trade wasn't after the last candle (open position at end)
            if str(int(t.closed_at)) in price_dict[key]['indexes']:
                # add price change%
                buy_y.append(
                    price_dict[key]['prices'][price_dict[key]['indexes'][str(int(t.closed_at))]]
                )
                # add datetime
                buy_x.append(datetime.fromtimestamp(t.closed_at / 1000))

            # Sell
            if str(int(t.opened_at)) in price_dict[key]['indexes']:
                # add price change%
                sell_y.append(
                    price_dict[key]['prices'][price_dict[key]['indexes'][str(int(t.opened_at))]]
                )
                # add datetime
                sell_x.append(datetime.fromtimestamp(t.opened_at / 1000))

    plt.plot(buy_x, np.array(buy_y) * 0.99, '^', color='blue', markersize=7)
    plt.plot(sell_x, np.array(sell_y) * 1.01, 'v', color='red', markersize=7)

    plt.xlabel('date')
    plt.ylabel('price change %')
    plt.title('Asset Daily Return')
    plt.legend(loc='upper left')

    # store final result
    # make sure directories exist
    os.makedirs('./storage/charts', exist_ok=True)
    file_path = f'storage/charts/{jh.get_session_id()}.png'.replace(":", "-")
    plt.savefig(file_path)

    return file_path



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, np_shift, slice_candles


def vpt(candles: np.ndarray, source_type: str = "close", sequential: bool = False) -> Union[float, np.ndarray]:
    """
    Volume Price Trend (VPT)

    :param candles: np.ndarray
    :param source_type: str - default: "close"
    :param sequential: bool - default: False

    :return: float | np.ndarray
    """
    candles = slice_candles(candles, sequential)

    source = get_candle_source(candles, source_type=source_type)
    vpt_val = (candles[:, 5] * ((source - np_shift(source, 1, fill_value=np.nan)) / np_shift(source, 1, fill_value=np.nan)))
    res = np_shift(vpt_val, 1, fill_value=np.nan) + vpt_val

    return res if sequential else res[-1]



import jesse.services.selectors as selectors
from jesse.config import config, reset_config
from jesse.enums import exchanges
from jesse.store import store


def set_up():
    reset_config()
    config['app']['considering_exchanges'] = [exchanges.SANDBOX]
    config['app']['trading_exchanges'] = [exchanges.SANDBOX]
    config['env']['exchanges'][exchanges.SANDBOX]['assets'] = [
        {'asset': 'USDT', 'balance': 2000}
    ]
    store.reset()


def test_have_correct_exchanges_in_store_after_creating_store():
    set_up()

    e = selectors.get_exchange(exchanges.SANDBOX)
    assert len(store.exchanges.storage) == 1
    assert e.assets['USDT'] == 2000



from jesse.strategies import Strategy


# test_modifying_take_profit_after_opening_position
class Test12(Strategy):
    def should_long(self):
        return self.price < 7

    def should_short(self):
        return False

    def go_long(self):
        qty = 1.5
        self.buy = qty, 7
        self.stop_loss = qty, 5
        self.take_profit = [
            (0.5, 11),
            (0.5, 13),
            (0.5, 15)
        ]

    def go_short(self):
        pass

    def should_cancel(self):
        return False

    def filters(self):
        return []

    def update_position(self):
        if self.price == 10:
            self.take_profit = self.position.qty, 16
